#
# Copyright (c) 2022-2023, NVIDIA CORPORATION. All rights reserved.
#
from urllib.parse import urljoin, urlencode
from typing import TypeVar, Type, Any, Dict

import requests

from aistore.sdk.const import (
    JSON_CONTENT_TYPE,
    HEADER_USER_AGENT,
    USER_AGENT_BASE,
    HEADER_CONTENT_TYPE,
    HEADERS_KW,
)
from aistore.sdk.utils import handle_errors, decode_response
from aistore.version import __version__ as sdk_version

T = TypeVar("T")


# pylint: disable=unused-variable
class RequestClient:
    """
    Internal client for buckets, objects, jobs, etc. to use for making requests to an AIS cluster

    Args:
        endpoint (str): AIStore endpoint
    """

    def __init__(self, endpoint: str):
        self._endpoint = endpoint
        self._base_url = urljoin(endpoint, "v1")
        self._session = requests.session()

    @property
    def base_url(self):
        """
        Returns: AIS cluster base url
        """
        return self._base_url

    @property
    def endpoint(self):
        """
        Returns: AIS cluster endpoint
        """
        return self._endpoint

    @property
    def session(self):
        """
        Returns: Active request session
        """
        return self._session

    def request_deserialize(
        self, method: str, path: str, res_model: Type[T], **kwargs
    ) -> T:
        """
        Make a request to the AIS cluster and deserialize the response to a defined type
        Args:
            method (str): HTTP method, e.g. POST, GET, PUT, DELETE
            path (str): URL path to call
            res_model (Type[T]): Resulting type to which the response should be deserialized
            **kwargs (optional): Optional keyword arguments to pass with the call to request

        Returns:
            Parsed result of the call to the API, as res_model
        """
        resp = self.request(method, path, **kwargs)
        return decode_response(res_model, resp)

    def request(self, method: str, path: str, **kwargs) -> requests.Response:
        """
        Make a request to the AIS cluster
        Args:
            method (str): HTTP method, e.g. POST, GET, PUT, DELETE
            path (str): URL path to call
            **kwargs (optional): Optional keyword arguments to pass with the call to request

        Returns:
            Raw response from the API
        """
        url = f"{self._base_url}/{path.lstrip('/')}"
        if HEADERS_KW not in kwargs:
            kwargs[HEADERS_KW] = {}
        headers = kwargs.get(HEADERS_KW, {})
        headers[HEADER_CONTENT_TYPE] = JSON_CONTENT_TYPE
        headers[HEADER_USER_AGENT] = f"{USER_AGENT_BASE}/{sdk_version}"
        resp = self._session.request(
            method,
            url,
            **kwargs,
        )
        if resp.status_code < 200 or resp.status_code >= 300:
            handle_errors(resp)
        return resp

    def get_full_url(self, path: str, params: Dict[str, Any]):
        """
        Get the full URL to the path on the cluster with the parameters given

        Args:
            path: Path on the cluster
            params: Query parameters to include

        Returns:
            URL including cluster base url and parameters

        """
        return f"{self._base_url}/{path.lstrip('/')}?{urlencode(params)}"



"""
AIS Plugin for PyTorch

PyTorch Dataset and DataLoader for AIS.

Copyright (c) 2022-2023, NVIDIA CORPORATION. All rights reserved.
"""

from typing import List, Union
from torch.utils.data import Dataset

from aistore.sdk import Client
from aistore.pytorch.utils import list_objects_info

# pylint: disable=too-few-public-methods


class AISBaseClass:
    """
    A base class for creating AIS Datasets for PyTorch

    Args:
        client_url(str): AIS endpoint URL
        urls_list(str or List[str]): single or list of url prefixes to load data
    """

    def __init__(self, client_url: str, urls_list: Union[str, List[str]]) -> None:
        self.client = Client(client_url)
        if isinstance(urls_list, str):
            urls_list = [urls_list]
        self._object_info = {}
        self._object_info = list_objects_info(self.client, urls_list)


# pylint: disable=unused-variable


class AISDataset(AISBaseClass, Dataset):
    """
    A map-style dataset for objects in AIS.
    If `etl_name` is provided, that ETL must already exist on the AIStore cluster

    Args:
        client_url (str): AIS endpoint URL
        urls_list (str or List[str]): single or list of url prefixes to load data
        etl_name (str, optional): Optional etl on the AIS cluster to apply to each object
    """

    def __init__(
        self, client_url: str, urls_list: Union[str, List[str]], etl_name=None
    ):
        AISBaseClass.__init__(self, client_url, urls_list)
        self.etl_name = etl_name

    def __len__(self):
        return len(self._object_info)

    def __getitem__(self, index: int):
        object_name = self._object_info[index]["object"]
        obj = (
            self.client.bucket(
                bck_name=self._object_info[index]["bck_name"],
                provider=self._object_info[index]["provider"],
            )
            .object(obj_name=object_name)
            .get(etl_name=self.etl_name)
            .read_all()
        )
        return object_name, obj



from aistore.pytorch.aisio import (
    AISFileListerIterDataPipe as AISFileLister,
    AISFileLoaderIterDataPipe as AISFileLoader,
    AISSourceLister,
)
from aistore.pytorch.dataset import AISDataset



#
# Copyright (c) 2022-2023, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable
from typing import List, Optional

from aistore.sdk.const import (
    HTTP_METHOD_GET,
    ACT_LIST,
    PROVIDER_AIS,
    QPARAM_WHAT,
    QPARAM_PRIMARY_READY_REB,
    QPARAM_PROVIDER,
    WHAT_SMAP,
    URL_PATH_BUCKETS,
    URL_PATH_HEALTH,
    URL_PATH_DAEMON,
    URL_PATH_CLUSTER,
    WHAT_ALL_XACT_STATUS,
    WHAT_ALL_RUNNING_STATUS,
    URL_PATH_ETL,
)

from aistore.sdk.types import BucketModel, JobStatus, JobQuery, ETLInfo
from aistore.sdk.request_client import RequestClient
from aistore.sdk.types import ActionMsg, Smap


# pylint: disable=unused-variable
class Cluster:
    """
    A class representing a cluster bound to an AIS client.
    """

    # pylint: disable=duplicate-code
    def __init__(self, client: RequestClient):
        self._client = client

    @property
    def client(self):
        """Client this cluster uses to make requests"""
        return self._client

    def get_info(self) -> Smap:
        """
        Returns state of AIS cluster, including the detailed information about its nodes.

        Returns:
            aistore.sdk.types.Smap: Smap containing cluster information

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_DAEMON,
            res_model=Smap,
            params={QPARAM_WHAT: WHAT_SMAP},
        )

    def list_buckets(self, provider: str = PROVIDER_AIS):
        """
        Returns list of buckets in AIStore cluster.

        Args:
            provider (str, optional): Name of bucket provider, one of "ais", "aws", "gcp", "az", "hdfs" or "ht".
            Defaults to "ais". Empty provider returns buckets of all providers.

        Returns:
            List[BucketModel]: A list of buckets

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = {QPARAM_PROVIDER: provider}
        action = ActionMsg(action=ACT_LIST).dict()

        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_BUCKETS,
            res_model=List[BucketModel],
            json=action,
            params=params,
        )

    def list_jobs_status(self, job_kind="", target_id="") -> List[JobStatus]:
        """
        List the status of jobs on the cluster

        Args:
            job_kind (str, optional): Only show jobs of a particular type
            target_id (str, optional): Limit to jobs on a specific target node

        Returns:
            List of JobStatus objects
        """
        res = self._client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_CLUSTER,
            res_model=Optional[List[JobStatus]],
            json=JobQuery(kind=job_kind, target=target_id).as_dict(),
            params={QPARAM_WHAT: WHAT_ALL_XACT_STATUS},
        )
        if res is None:
            return []
        return res

    def list_running_jobs(self, job_kind="", target_id="") -> List[str]:
        """
        List the currently running jobs on the cluster

        Args:
            job_kind (str, optional): Only show jobs of a particular type
            target_id (str, optional): Limit to jobs on a specific target node

        Returns:
            List of jobs in the format job_kind[job_id]
        """
        return self._client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_CLUSTER,
            res_model=List[str],
            json=JobQuery(kind=job_kind, target=target_id, active=True).as_dict(),
            params={QPARAM_WHAT: WHAT_ALL_RUNNING_STATUS},
        )

    def list_running_etls(self) -> List[ETLInfo]:
        """
        Lists all running ETLs.

        Note: Does not list ETLs that have been stopped or deleted.

        Returns:
            List[ETLInfo]: A list of details on running ETLs
        """
        return self._client.request_deserialize(
            HTTP_METHOD_GET, path=URL_PATH_ETL, res_model=List[ETLInfo]
        )

    def is_aistore_running(self) -> bool:
        """
        Checks if cluster is ready or still setting up.

        Returns:
            bool: True if cluster is ready, or false if cluster is still setting up
        """

        # compare with AIS Go API (api/cluster.go) for additional supported options
        params = {QPARAM_PRIMARY_READY_REB: "true"}
        try:
            resp = self.client.request(
                HTTP_METHOD_GET, path=URL_PATH_HEALTH, params=params
            )
            return resp.ok
        except Exception:
            return False



#
# Copyright (c) 2023, NVIDIA CORPORATION. All rights reserved.
#
import logging
from typing import List, Iterable

from aistore.sdk.ais_source import AISSource
from aistore.sdk.const import (
    HTTP_METHOD_DELETE,
    HTTP_METHOD_POST,
    HTTP_METHOD_PUT,
    ACT_DELETE_OBJECTS,
    ACT_PREFETCH_OBJECTS,
    ACT_EVICT_OBJECTS,
    ACT_COPY_OBJECTS,
    ACT_TRANSFORM_OBJECTS,
    ACT_ARCHIVE_OBJECTS,
)
from aistore.sdk.etl_const import DEFAULT_ETL_TIMEOUT
from aistore.sdk.multiobj.object_names import ObjectNames
from aistore.sdk.multiobj.object_range import ObjectRange
from aistore.sdk.multiobj.object_template import ObjectTemplate
from aistore.sdk.types import (
    TCMultiObj,
    CopyBckMsg,
    TransformBckMsg,
    TCBckMsg,
    ArchiveMultiObj,
)


# pylint: disable=unused-variable
class ObjectGroup(AISSource):
    """
    A class representing multiple objects within the same bucket. Only one of obj_names, obj_range, or obj_template
     should be provided.

    Args:
        bck (Bucket): Bucket the objects belong to
        obj_names (list[str], optional): List of object names to include in this collection
        obj_range (ObjectRange, optional): Range defining which object names in the bucket should be included
        obj_template (str, optional): String argument to pass as template value directly to api
    """

    def __init__(
        self,
        bck: "Bucket",
        obj_names: list = None,
        obj_range: ObjectRange = None,
        obj_template: str = None,
    ):
        self.bck = bck
        num_args = sum(
            1 if x is not None else 0 for x in [obj_names, obj_range, obj_template]
        )
        if num_args != 1:
            raise ValueError(
                "ObjectGroup accepts one and only one of: obj_names, obj_range, or obj_template"
            )
        if obj_range and not isinstance(obj_range, ObjectRange):
            raise TypeError("obj_range must be of type ObjectRange")

        if obj_range:
            self._obj_collection = obj_range
        elif obj_names:
            self._obj_collection = ObjectNames(obj_names)
        else:
            self._obj_collection = ObjectTemplate(obj_template)

    def list_urls(self, prefix: str = "", etl_name: str = None) -> Iterable[str]:
        """
            Get an iterator of the full URL for every object in this group
        Args:
            prefix (str, optional): Limit objects selected by a given string prefix
            etl_name (str, optional): ETL to include in URLs

        Returns:
            Iterator of all object URLs in the group
        """
        for obj_name in self._obj_collection:
            yield self.bck.object(obj_name).get_url(etl_name=etl_name)

    def delete(self):
        """
        Deletes a list or range of objects in a bucket

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        """

        return self.bck.make_request(
            HTTP_METHOD_DELETE,
            ACT_DELETE_OBJECTS,
            value=self._obj_collection.get_value(),
        ).text

    def evict(self):
        """
        Evicts a list or range of objects in a bucket so that they are no longer cached in AIS
        NOTE: only Cloud buckets can be evicted.

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        """
        self.bck.verify_cloud_bucket()
        return self.bck.make_request(
            HTTP_METHOD_DELETE,
            ACT_EVICT_OBJECTS,
            value=self._obj_collection.get_value(),
        ).text

    def prefetch(self):
        """
        Prefetches a list or range of objects in a bucket so that they are cached in AIS
        NOTE: only Cloud buckets can be prefetched.

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        """
        self.bck.verify_cloud_bucket()
        return self.bck.make_request(
            HTTP_METHOD_POST,
            ACT_PREFETCH_OBJECTS,
            value=self._obj_collection.get_value(),
        ).text

    # pylint: disable=too-many-arguments
    def copy(
        self,
        to_bck: "Bucket",
        prepend: str = "",
        continue_on_error: bool = False,
        dry_run: bool = False,
        force: bool = False,
    ):
        """
        Copies a list or range of objects in a bucket

        Args:
            to_bck (Bucket): Destination bucket
            prepend (str, optional): Value to prepend to the name of copied objects
            continue_on_error (bool, optional): Whether to continue if there is an error copying a single object
            dry_run (bool, optional): Skip performing the copy and just log the intended actions
            force (bool, optional): Force this job to run over others in case it conflicts
                (see "limited coexistence" and xact/xreg/xreg.go)

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        """
        if dry_run:
            logger = logging.getLogger(f"{__name__}.copy")
            logger.info(
                "Copy dry-run. Running with dry_run=False will copy the following objects from bucket '%s' to '%s': %s",
                f"{self.bck.get_path()}",
                f"{to_bck.get_path()}",
                list(self._obj_collection),
            )
        copy_msg = CopyBckMsg(prepend=prepend, dry_run=dry_run, force=force)

        value = TCMultiObj(
            to_bck=to_bck.as_model(),
            tc_msg=TCBckMsg(copy_msg=copy_msg),
            object_selection=self._obj_collection.get_value(),
            continue_on_err=continue_on_error,
        ).as_dict()
        return self.bck.make_request(
            HTTP_METHOD_POST, ACT_COPY_OBJECTS, value=value
        ).text

    # pylint: disable=too-many-arguments
    def transform(
        self,
        to_bck: "Bucket",
        etl_name: str,
        timeout: str = DEFAULT_ETL_TIMEOUT,
        prepend: str = "",
        continue_on_error: bool = False,
        dry_run: bool = False,
        force: bool = False,
    ):
        """
        Performs ETL operation on a list or range of objects in a bucket, placing the results in the destination bucket

        Args:
            to_bck (Bucket): Destination bucket
            etl_name (str): Name of existing ETL to apply
            timeout (str): Timeout of the ETL job (e.g. 5m for 5 minutes)
            prepend (str, optional): Value to prepend to the name of resulting transformed objects
            continue_on_error (bool, optional): Whether to continue if there is an error transforming a single object
            dry_run (bool, optional): Skip performing the transform and just log the intended actions
            force (bool, optional): Force this job to run over others in case it conflicts
                (see "limited coexistence" and xact/xreg/xreg.go)

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        """
        if dry_run:
            logger = logging.getLogger(f"{__name__}.transform")
            logger.info(
                "Transform dry-run. Running with dry_run=False will apply ETL '%s' to objects %s",
                etl_name,
                list(self._obj_collection),
            )

        copy_msg = CopyBckMsg(prepend=prepend, dry_run=dry_run, force=force)
        transform_msg = TransformBckMsg(etl_name=etl_name, timeout=timeout)
        value = TCMultiObj(
            to_bck=to_bck.as_model(),
            tc_msg=TCBckMsg(transform_msg=transform_msg, copy_msg=copy_msg),
            object_selection=self._obj_collection.get_value(),
            continue_on_err=continue_on_error,
        ).as_dict()
        return self.bck.make_request(
            HTTP_METHOD_POST, ACT_TRANSFORM_OBJECTS, value=value
        ).text

    def archive(
        self,
        archive_name: str,
        mime: str = "",
        to_bck: "Bucket" = None,
        include_source_name: bool = False,
        allow_append: bool = False,
        continue_on_err: bool = False,
    ):
        """
        Create or append to an archive

        Args:
            archive_name (str): Name of archive to create or append
            mime (str, optional): MIME type of the content
            to_bck (Bucket, optional): Destination bucket, defaults to current bucket
            include_source_name (bool, optional): Include the source bucket name in the archived objects' names
            allow_append (bool, optional): Allow appending to an existing archive
            continue_on_err (bool, optional): Whether to continue if there is an error archiving a single object

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        """
        val = ArchiveMultiObj(
            object_selection=self._obj_collection.get_value(),
            archive_name=archive_name,
            mime=mime,
            to_bck=to_bck.as_model() if to_bck else self.bck.as_model(),
            include_source_name=include_source_name,
            allow_append=allow_append,
            continue_on_err=continue_on_err,
        ).as_dict()
        return self.bck.make_request(
            HTTP_METHOD_PUT, ACT_ARCHIVE_OBJECTS, value=val
        ).text

    def list_names(self) -> List[str]:
        """
        List all the object names included in this group of objects

        Returns:
            List of object names

        """
        return list(self._obj_collection)



#
# Copyright (c) 2023, NVIDIA CORPORATION. All rights reserved.
#
from abc import ABC, abstractmethod
from typing import Iterable


# pylint: disable=too-few-public-methods
class AISSource(ABC):
    """
    Interface for all AIS class types providing access to AIS objects via URLs
    """

    @abstractmethod
    def list_urls(self, prefix: str = "", etl_name: str = None) -> Iterable[str]:
        """
        Get an iterable of urls to reference the objects contained in this source (bucket, group, etc.)
        Args:
            prefix (str, optional): Only include objects with names matching this prefix
            etl_name (str, optional): Apply an ETL when retrieving object contents

        Returns:
            Iterable over selected object URLS
        """



#
# Copyright (c) 2022-2023, NVIDIA CORPORATION. All rights reserved.
#
import sys

import base64
from typing import Callable, List

import cloudpickle

from aistore.sdk.const import (
    HTTP_METHOD_DELETE,
    HTTP_METHOD_GET,
    HTTP_METHOD_POST,
    HTTP_METHOD_PUT,
    URL_PATH_ETL,
    UTF_ENCODING,
)
from aistore.sdk.etl_const import (
    ETL_SUPPORTED_PYTHON_VERSIONS,
    DEFAULT_ETL_RUNTIME,
    DEFAULT_ETL_COMM,
    DEFAULT_ETL_TIMEOUT,
    ETL_COMM_SPEC,
    ETL_COMM_CODE,
    CODE_TEMPLATE,
)

from aistore.sdk.types import ETLDetails, InitCodeETLArgs, InitSpecETLArgs


def _get_default_runtime():
    """
    Determines etl runtime to use if not specified
    Returns:
        String of runtime
    """
    version = f"{sys.version_info.major}.{sys.version_info.minor}"
    if version in ETL_SUPPORTED_PYTHON_VERSIONS:
        return f"python{version}v2"
    return DEFAULT_ETL_RUNTIME


# pylint: disable=unused-variable
def _validate_comm_type(given: str, valid: List[str]):
    if given not in valid:
        valid_str = ", ".join(valid)
        raise ValueError(f"communication_type should be one of: {valid_str}")


class Etl:
    """
    A class containing ETL-related functions.
    """

    def __init__(self, client: "Client", name: str):
        self._client = client
        self._name = name

    @property
    def name(self) -> str:
        """Name of the ETL"""
        return self._name

    def init_spec(
        self,
        template: str,
        communication_type: str = DEFAULT_ETL_COMM,
        timeout: str = DEFAULT_ETL_TIMEOUT,
    ) -> str:
        """
        Initializes ETL based on Kubernetes pod spec template.

        Args:
            template (str): Kubernetes pod spec template
                Existing templates can be found at `sdk.etl_templates`
                For more information visit: https://github.com/NVIDIA/ais-etl/tree/master/transformers
            communication_type (str): Communication type of the ETL (options: hpull, hrev, hpush)
            timeout (str): Timeout of the ETL job (e.g. 5m for 5 minutes)
        Returns:
            Job ID string associated with this ETL
        """
        _validate_comm_type(communication_type, ETL_COMM_SPEC)

        # spec
        spec_encoded = base64.b64encode(template.encode(UTF_ENCODING)).decode(
            UTF_ENCODING
        )

        value = InitSpecETLArgs(
            spec=spec_encoded,
            etl_name=self._name,
            communication_type=communication_type,
            timeout=timeout,
        ).as_dict()

        return self._client.request(HTTP_METHOD_PUT, path=URL_PATH_ETL, json=value).text

    # pylint: disable=too-many-arguments
    def init_code(
        self,
        transform: Callable,
        dependencies: List[str] = None,
        preimported_modules: List[str] = None,
        runtime: str = _get_default_runtime(),
        communication_type: str = DEFAULT_ETL_COMM,
        timeout: str = DEFAULT_ETL_TIMEOUT,
        chunk_size: int = None,
        transform_url: bool = False,
    ) -> str:
        """
        Initializes ETL based on the provided source code.

        Args:
            transform (Callable): Transform function of the ETL
            dependencies (list[str]): Python dependencies to install
            preimported_modules (list[str]): Modules to import before running the transform function. This can
             be necessary in cases where the modules used both attempt to import each other circularly
            runtime (str): [optional, default= V2 implementation of the current python version if supported, else
                python3.8v2] Runtime environment of the ETL [choose from: python3.8v2, python3.10v2, python3.11v2]
                (see ext/etl/runtime/all.go)
            communication_type (str): [optional, default="hpush"] Communication type of the ETL (options: hpull, hrev,
                hpush, io)
            timeout (str): [optional, default="5m"] Timeout of the ETL job (e.g. 5m for 5 minutes)
            chunk_size (int): Chunk size in bytes if transform function in streaming data.
                (whole object is read by default)
            transform_url (optional, bool): If True, the runtime will provide the transform function with the URL to the
             object on the target rather than the raw bytes read from the object
        Returns:
            Job ID string associated with this ETL
        """
        _validate_comm_type(communication_type, ETL_COMM_CODE)

        # code functions to call
        functions = {
            "transform": "transform",
        }

        value = InitCodeETLArgs(
            etl_name=self._name,
            runtime=runtime,
            communication_type=communication_type,
            timeout=timeout,
            dependencies=self._encode_dependencies(dependencies),
            functions=functions,
            code=self._encode_transform(
                transform, preimported_modules, communication_type
            ),
            chunk_size=chunk_size,
            transform_url=transform_url,
        ).as_dict()

        return self._client.request(
            HTTP_METHOD_PUT,
            path=URL_PATH_ETL,
            json=value,
        ).text

    def view(self) -> ETLDetails:
        """
        View ETL details

        Returns:
            ETLDetails: details of the ETL
        """
        resp = self._client.request_deserialize(
            HTTP_METHOD_GET, path=f"{URL_PATH_ETL}/{self._name}", res_model=ETLDetails
        )
        return resp

    def start(self):
        """
        Resumes a stopped ETL with given ETL name.

        Note: Deleted ETLs cannot be started.
        """
        self._client.request(
            HTTP_METHOD_POST, path=f"{URL_PATH_ETL}/{self._name}/start"
        )

    def stop(self):
        """
        Stops ETL. Stops (but does not delete) all the pods created by Kubernetes for this ETL and
        terminates any transforms.
        """
        self._client.request(HTTP_METHOD_POST, path=f"{URL_PATH_ETL}/{self._name}/stop")

    def delete(self):
        """
        Delete ETL. Deletes pods created by Kubernetes for this ETL and specifications for this ETL
        in Kubernetes.

        Note: Running ETLs cannot be deleted.
        """
        self._client.request(HTTP_METHOD_DELETE, path=f"{URL_PATH_ETL}/{self._name}")

    @staticmethod
    def _encode_transform(
        transform: Callable,
        preimported_modules: List[str] = None,
        comm_type: str = None,
    ):
        transform = base64.b64encode(cloudpickle.dumps(transform)).decode(UTF_ENCODING)

        io_comm_context = "transform()" if comm_type == "io" else ""
        modules = preimported_modules if preimported_modules else []
        template = CODE_TEMPLATE.format(modules, transform, io_comm_context).encode(
            UTF_ENCODING
        )
        return base64.b64encode(template).decode(UTF_ENCODING)

    @staticmethod
    def _encode_dependencies(dependencies: List[str]):
        if dependencies is None:
            dependencies = []
        dependencies.append("cloudpickle==2.2.0")
        deps = "\n".join(dependencies).encode(UTF_ENCODING)
        return base64.b64encode(deps).decode(UTF_ENCODING)
